/*
	File:		SunSkyConstants.cc

	Function:	

    Author:     Brian Smits (bes@phoenix.cs.utah.edu),
    			Some minor modifications by Andrew Willmott (ajw@cs.cmu.edu)

	Copyright:	(c) 1999, Brian Smits
*/


#ifdef __GNUC__
// GNU math inlines are disabled for gcc 2.7 and below because of "problems".
// unfortunately 2.8 also has problems with calls to pow() in this file,
// so we turn them off for it, too. sigh. ajw
#if ((__GNUC__ == 2 && __GNUC_MINOR__ <= 8))
#undef __NO_MATH_INLINES
#define __NO_MATH_INLINES
#endif
#endif

#include "SunSky.h"

// 300-800 at 10 nm
static Float n2_1Amplitudes[51] =
{
    3.4012e-7, 3.3702e-7, 3.3425e-7, 3.3177e-7, 3.2955e-7,
    3.2754e-7, 3.2572e-7, 3.2406e-7, 3.2255e-7, 3.2116e-7,

    3.1989e-7, 3.1873e-7, 3.1765e-7, 3.1665e-7, 3.1572e-7,
    3.1486e-7, 3.1407e-7, 3.1332e-7, 3.1263e-7, 3.1198e-7,

    3.1137e-7, 3.1080e-7, 3.1026e-7, 3.0976e-7, 3.0928e-7,
    3.0884e-7, 3.0841e-7, 3.0801e-7, 3.0763e-7, 3.0728e-7,

    3.0694e-7, 3.0661e-7, 3.0631e-7, 3.0602e-7, 3.0574e-7,
    3.0547e-7, 3.0522e-7, 3.0498e-7, 3.0475e-7, 3.0453e-7,

    3.0432e-7, 3.0412e-7, 3.0393e-7, 3.0375e-7, 3.0357e-7,
    3.0340e-7, 3.0324e-7, 3.0309e-7, 3.0293e-7, 3.0279e-7,

    3.0265e-7
};

// 300-800 at 1nm
static Float KAmplitudes[501] =
{
#include "K.spectrum"
};

static Float K25Amplitudes[501] =
{
#include "K25.spectrum"
};

//300- 790 at 10nm  (for each of 1801 tenths of a degree
static Float netaLambdaTable[50][1801] =
{
#include "Neta.Table"
};

static Spectrum netaTable[1801];

static GCLReal const Anis_cor = 1.06;  	// Correction for molecular anisotropy.
static GCLReal const N = 2.545e25; 		// Number of molecules per unit volume.

static void InitNetaTable()
{
    Float data[50];

    for (Int i = 0; i < 1801; i++)
    {
        for (Int j = 0; j < 50; j++)
            data[j] = netaLambdaTable[j][i];
        netaTable[i] = RegularSpectralCurve(data, 300, 790, 50);
    }
}

/* might need to be converted because the above functions use a
   different system ... */
  
void SunSky::CreateConstants()
{
    InitNetaTable();

    GCLReal	lambda;
    Float	b_m[101], b_p[101];
    Float	b_m_ang_prefix[101];
    Float	b_p_ang_prefix[101];

    Int i;
    GCLReal lambdasi, Nlambda4;
    GCLReal n2_1;
    GCLReal K;
    GCLReal c = (0.06544204545455 * turbidity -0.06509886363636) * 1e-15;

    RegularSpectralCurve KCurve(KAmplitudes, 300, 800, 501);
    RegularSpectralCurve K25Curve(K25Amplitudes, 300, 800, 501);
    RegularSpectralCurve n2_1Curve(n2_1Amplitudes, 300, 800, 51);

    for (lambda = 300.0, i = 0; lambda <= 800.0; lambda += 5.0, i++)
    {
        lambdasi = lambda * 1e-9;  /* Converstion to SI units */

        Nlambda4 = N * sqr(sqr(lambdasi));

        /* Rayleigh total scattering coefficient */
        /* (4-48), p 199, MC */
        n2_1 = n2_1Curve.Sample(lambda);
        b_m[i] = 8.0 * vl_pi * vl_pi * vl_pi * n2_1 * Anis_cor /
                 (3.0 * Nlambda4);


        /* Mie total scattering coefficient */
        /* (6-30). p 280; MC */
        if (V > 3.9)
        	K = KCurve.Sample(lambda); 		// V = 4.0;
        else
        	K = K25Curve.Sample(lambda);    // V = 2.5

        b_p[i] = 0.434*vl_pi*c*pow(GCLReal(2*vl_pi/lambdasi),V-2) * K *
                 pow(GCLReal(0.01),V-3);  /* Last term is unit correction for c */

        /* Rayleigh Angular scattering coefficient */
        /*  (4-55), p 200; MC */
        /* Needs to be multiplied by (1+0.9324cos^2(theta)) to get exact
           angular scattering constant */
        b_m_ang_prefix[i] = 2*vl_pi*vl_pi*n2_1*Anis_cor*0.7629/
                            (3*Nlambda4);


        /* Mie Angular scattering coefficient */
        /* (6-24), p 271; MC */
        /* Needs to be multiplied by neta(theta,lambda) to get exact
           angular scattering constant term */
        b_p_ang_prefix[i] = 0.217*c*pow(GCLReal(2*vl_pi/lambdasi),V-2)*
                            pow(GCLReal(0.01),V-3);  /* Last term is unit correction for c */
    }

    beta_m = RegularSpectralCurve(b_m, 300, 800, 101);
    beta_p = RegularSpectralCurve(b_p, 300, 800, 101);
    beta_m_ang_prefix = RegularSpectralCurve(b_m_ang_prefix, 300, 800, 101);
    beta_p_ang_prefix = RegularSpectralCurve(b_p_ang_prefix, 300, 800, 101);
}



Spectrum SunSky::GetNeta(GCLReal theta, GCLReal v) const
{
    theta = RadsToDegs(theta) * 10;

    GCLReal u = theta - (Int) theta;

    if ((theta < 0) || (theta > 1801))
    {
        cerr << "Error:getNeta() [theta outside range] theta=" << theta/10
        <<"degrees" << endl;
        return 0;
    }

    if (theta > 1800)
    	theta = 1800;

    if ((Int) theta == 1800)
        return  netaTable[1800];

    return (1.0 - u) * netaTable[(Int) theta] + u * netaTable[(Int) theta + 1];
}

/* All data lifted from MI */
/* Units are either [] or cm^-1. refer when in doubt MI */

// k_o Spectrum table from pg 127, MI.
static Float k_oWavelengths[64] =
{
    300, 305, 310, 315, 320,
    325, 330, 335, 340, 345,
    350, 355,

    445, 450, 455, 460, 465,
    470, 475, 480, 485, 490,
    495,

    500, 505, 510, 515, 520,
    525, 530, 535, 540, 545,
    550, 555, 560, 565, 570,
    575, 580, 585, 590, 595,

    600, 605, 610, 620, 630,
    640, 650, 660, 670, 680,
    690,

    700, 710, 720, 730, 740,
    750, 760, 770, 780, 790,
};

static Float k_oAmplitudes[65] = {
    10.0,
    4.8,
    2.7,
    1.35,
    .8,
    .380,
    .160,
    .075,
    .04,
    .019,
    .007,
    .0,

    .003,
    .003,
    .004,
    .006,
    .008,
    .009,
    .012,
    .014,
    .017,
    .021,
    .025,

    .03,
    .035,
    .04,
    .045,
    .048,
    .057,
    .063,
    .07,
    .075,
    .08,
    .085,
    .095,
    .103,
    .110,
    .12,
    .122,
    .12,
    .118,
    .115,
    .12,

    .125,
    .130,
    .12,
    .105,
    .09,
    .079,
    .067,
    .057,
    .048,
    .036,
    .028,

    .023,
    .018,
    .014,
    .011,
    .010,
    .009,
    .007,
    .004,
    .0,
    .0
};


// k_g Spectrum table from pg 130, MI.
static Float k_gWavelengths[4] = {
    759,
    760,
    770,
    771
};

static Float k_gAmplitudes[4] = {
    0,
    3.0,
    0.210,
    0
};

// k_wa Spectrum table from pg 130, MI.
static Float k_waWavelengths[13] = {
    689,
    690,
    700,
    710,
    720,
    730,
    740,
    750,
    760,
    770,
    780,
    790,
    800
};

static Float k_waAmplitudes[13] = {
    0,
    0.160e-1,
    0.240e-1,
    0.125e-1,
    0.100e+1,
    0.870,
    0.610e-1,
    0.100e-2,
    0.100e-4,
    0.100e-4,
    0.600e-3,
    0.175e-1,
    0.360e-1
};

// Taken from Appendix C, MI   300-798 every 2nm
/*
  static Float solAmplitudes[250] = {
  0.99, 1.10, 1.06, 1.09, 1.22, 1.10, 1.30, 1.36, 1.42, 1.44,
  1.43, 1.56, 1.36, 1.39, 2.12, 1.99, 2.00, 1.90, 1.87, 1.61,
  1.90, 1.94, 1.89, 1.66, 1.81, 1.80, 2.02, 2.04, 2.27, 1.78,
  1.72, 2.04, 2.16, 2.19, 2.46, 2.35, 2.31, 1.89, 2.07, 2.62,
  2.51, 2.33, 1.44, 2.04, 1.93, 2.24, 2.56, 1.49, 2.53, 1.72,

  3.21, 3.60, 3.36, 3.27, 3.25, 3.39, 3.45, 3.52, 3.45, 3.56,
  3.32, 3.67, 3.29, 3.50, 3.25, 3.05, 2.92, 3.31, 3.54, 3.70,
  3.38, 3.64, 3.90, 3.78, 3.92, 4.03, 4.37, 3.89, 4.06, 4.13,
  3.98, 4.15, 4.17, 3.96, 3.91, 4.02, 3.90, 4.04, 4.03, 4.03,
  4.12, 4.10, 4.01, 3.83, 3.54, 3.88, 3.80, 3.84, 3.95, 3.98,

  3.82, 3.63, 3.84, 3.93, 3.85, 3.88, 3.93, 3.74, 3.77, 3.40,
  3.45, 3.76, 3.84, 3.88, 3.56, 3.84, 3.95, 3.76, 3.92, 3.81,
  3.76, 3.67, 3.78, 3.81, 3.76, 3.84, 3.77, 3.77, 3.82, 3.67,
  3.67, 3.69, 3.76, 3.67, 3.70, 3.70, 3.57, 3.74, 3.72, 3.61,
  3.64, 3.72, 3.69, 3.62, 3.61, 3.35, 3.62, 3.54, 3.57, 3.60,

  3.44, 3.44, 3.55, 3.47, 3.52, 3.39, 3.47, 3.42, 3.42, 3.36,
  3.49, 3.37, 3.38, 3.24, 3.40, 3.32, 3.23, 3.33, 3.32, 3.32,
  3.31, 3.18, 3.28, 3.26, 3.20, 3.14, 3.25, 3.21, 2.97, 2.81,
  3.07, 3.15, 3.16, 3.13, 3.07, 3.04, 2.98, 2.93, 3.03, 2.97,
  2.94, 2.95, 2.93, 2.89, 2.79, 2.80, 2.86, 2.85, 2.92, 2.92,

  2.87, 2.84, 2.81, 2.83, 2.76, 2.77, 2.84, 2.78, 2.75, 2.74,
  2.71, 2.71, 2.81, 2.79, 2.73, 2.70, 2.72, 2.71, 2.71, 2.69,
  2.61, 2.59, 2.57, 2.58, 2.58, 2.57, 2.52, 2.50, 2.50, 2.48,
  2.45, 2.44, 2.44, 2.41, 2.41, 2.38, 2.37, 2.36, 2.35, 2.34,
  2.40, 2.38, 2.35, 2.35, 2.33, 2.30, 2.27, 2.26, 2.33, 2.31
};
*/

// 380-750 by 10nm
static Float solAmplitudes[38] =
{
    165.5, 162.3, 211.2, 258.8, 258.2,
    242.3, 267.6, 296.6, 305.4, 300.6,
    306.6, 288.3, 287.1, 278.2, 271.0,
    272.3, 263.6, 255.0, 250.6, 253.1,
    253.5, 251.3, 246.3, 241.7, 236.8,
    232.1, 228.2, 223.4, 219.7, 215.3,
    211.0, 207.3, 202.4, 198.7, 194.3,
    190.7, 186.3, 182.6
};


/**********************************************************
// Sunlight Transmittance Functions
// 
// ********************************************************/

/* Most units not in SI system - For units, refer MI */

RegularSpectralCurve solCurve(solAmplitudes, 380, 750, 38);
IrregularSpectralCurve	k_oCurve(k_oAmplitudes, k_oWavelengths, 64);
IrregularSpectralCurve	k_gCurve(k_gAmplitudes, k_gWavelengths, 4);
IrregularSpectralCurve	k_waCurve(k_waAmplitudes, k_waWavelengths, 13);

Spectrum SunSky::ComputeAttenuatedSunlight(GCLReal theta, Int turbidity)
{
    static Float  	data[91];  // (800 - 350) / 5  + 1

    GCLReal beta = 0.04608365822050 * turbidity - 0.04586025928522;
    GCLReal tauR, tauA, tauO, tauG, tauWA;

    //    GCLReal m = 1.0/(cos(theta) + 0.15*pow(93.885-theta/vl_pi*180.0,-1.253));  
    // equivalent  (CHECK WITH PAPER)
    GCLReal		m = 1.0 / (cos(theta) + 0.000940 * pow(1.6386 - theta, -1.253));  
    // Relative Optical Mass

    Int 		i;
    GCLReal		lambda;

    for (i = 0, lambda = 350; i < 91; i++, lambda += 5)
    {
        // Rayleigh Scattering
        // Results agree with the graph (pg 115, MI) */
		tauR = exp( -m * 0.008735 * pow(lambda/1000, GCLReal(-4.08)));

        // Aerosal (water + dust) attenuation
        // beta - amount of aerosols present
        // alpha - ratio of small to large particle sizes. (0:4,usually 1.3)
        // Results agree with the graph (pg 121, MI)
        const GCLReal alpha = 1.3;
		tauA = exp(-m * beta * pow(lambda/1000, -alpha));  // lambda should be in um

        // Attenuation due to ozone absorption
        // lOzone - amount of ozone in cm(NTP)
        // Results agree with the graph (pg 128, MI)
        const GCLReal lOzone = .35;
        tauO = exp(-m * k_oCurve.Sample(lambda) * lOzone);

        // Attenuation due to mixed gases absorption
        // Results agree with the graph (pg 131, MI)
        tauG = exp(-1.41 * k_gCurve.Sample(lambda) * m
					/ pow(1.0 + 118.93 * k_gCurve.Sample(lambda) * m, 0.45));

        // Attenuation due to water vapor absorbtion
        // w - precipitable water vapor in centimeters (standard = 2)
        // Results agree with the graph (pg 132, MI)
		const GCLReal w = 2.0;

		tauWA = exp(-0.2385 * k_waCurve.Sample(lambda) * w * m 
					/ pow(1.0 + 20.07 * k_waCurve.Sample(lambda) * w * m, 0.45));

        data[i] = 106 * solCurve.Sample(lambda) * tauR * tauA * tauO * tauG * tauWA;
    }

    // Converts to Spectrum
	return(RegularSpectralCurve(data, 350, 800, 91));
}
